;;; integer-graph.scm   --  Jens Axel S?gaard

;;; THIS VERSION REPLACES THE OLD

; This version represent edges with the edge struct
; in order to associate information with edges such as weights.

(module integer-graphs-with-edges mzscheme
  
  (print-struct #t)
  
  (require (prefix ms: "monotone-set.scm")
           (prefix ns: "non-deterministic-set.scm")
           (prefix ps: "priority-set.scm")
           (prefix es: "equiv-set.scm")
           (prefix pq: "priority-queue.scm")
           "evector.scm"
  
           (prefix srfi: (lib "1.ss" "SRFI"))
           (lib "match.ss" "mzlib")
           (lib "list.ss" "mzlib"))
  
  (define-syntax thunk (syntax-rules () [(_ e ...) (lambda () e ...)]))
  
  (define (interval m n)
    (if (> m n)
        '()
        (cons m (interval (add1 m) n))))
  
  ;;; REPRESENTATION
  
  ; A NODE is a non-negative integer.
  
  ; A (abstract) EDGE is a pair of numbers.
  
  ; A GRAPH consists of a set of nodes (represented as a evector of nodes)
  ; and a set of edges represented using edge-lists. Each node v has an
  ; associated set (represented as an ordered list) of all nodes w such 
  ; that (v,w) is an edge.
  
  
  ;(require (lib "set.scm" "galore"))
  (require (planet "set.scm" ("soegaard" "galore.plt" 1 1)))
  
  (use-set (instantiate-set ordered-list-set@ integers@) is)   ; is stands for Integer Set
  
  (use-set (instantiate-set ordered-list-set@ equal-hash-order@) es) ; "es" stands for Edge Set
  
  
  (define-struct graph 
                 (oriented?        ; #t or #f
                  edge-lists       ; an evector of sets of edges
                  no-nodes)        ; the current number of nodes
                 (make-inspector)) ; for debug printing
  (define %make-graph make-graph)
  ; smart constructor
  (set! make-graph 
        (lambda (oriented?)
          (%make-graph oriented? (make-evector 0 es:empty 'automatic-expansion-on-set!) 0)))
  
  (define-struct edge (from to) (make-inspector))
  (define-struct (edge/weight edge) (weight) (make-inspector))

  
  
  (define (graph-unoriented? G)
    (not (graph-oriented? G)))
  
  
  (define (edge-list G n)
    ; return the edge set associated with the node n
    (evector-ref (graph-edge-lists G) n))
  
  (define (lookup-edge G from to)
    ; is (from,to) an edge in G?
    (ormap (match-lambda 
             [($ edge u v) (= to v)])
           (edge-list G from)))
  
  (define add-edge 
    ; AN EDGE CAN ONLY BE INSERTED ONCE,
    ; INSERTING TWICE CREATES TWO EDGES
    ; add (from,to) to the graph G (destructively)
    (case-lambda
      [(G from to)
       (let ([el (graph-edge-lists G)])
         (when (>= (max from to) (evector-length el))
           (set-evector-length! el (add1 (max from to))))
         (evector-set! el from (es:insert1 (make-edge from to) (evector-ref el from)))
         (when (graph-unoriented? G)
           (evector-set! el to (es:insert1 (make-edge to from) (evector-ref el to))))
         (set-graph-no-nodes! G (max (add1 from) (add1 to) (graph-no-nodes G))))]
      [(G from to weight)
       (let ([el (graph-edge-lists G)])
         (when (>= (max from to) (evector-length el))
           (set-evector-length! el (add1 (max from to))))
         (evector-set! el from (es:insert1 (make-edge/weight from to weight) (evector-ref el from)))
         (when (graph-unoriented? G)
           (evector-set! el to (es:insert1 (make-edge/weight to from weight) (evector-ref el to))))
         (set-graph-no-nodes! G (max (add1 from) (add1 to) (graph-no-nodes G))))]))
  
  
  (define-syntax for-each-node
    (syntax-rules (in)
      [(_ n in G body)
       (do ([n 0 (add1 n)])
         [(= n (evector-length (graph-edge-lists G))) (void)]
         body)]))
  
  (define-syntax for-each-edge
    (syntax-rules (from to in =)
      [(_ e = (u v w) from n in G body)
       (for-each (lambda (e)
                   (match e
                     [($ edge/weight u v w) body]))
                 (evector-ref (graph-edge-lists G) n))]
      [(_ (u v) from n in G body)
       (for-each (match-lambda [($ edge u v) body])
                 (evector-ref (graph-edge-lists G) n))]
      [(_ e from v in G body)
       (for-each (lambda (e) body)
                 (evector-ref (graph-edge-lists G) v))]
      [(_ (u v) to n in G body)
       (for-each (match-lambda [($ edge u v) body])
                 (in-edges G n))]
      [(_ e to v in G body)
       (for-each (lambda (e) body)
                 (in-edges G v))]
      [(_ (u v w) in G body)
       (for-each (match-lambda [($ edge/weight u v w) body])
                 (edges G))]
      [(_ e = (u v w) in G body)
       (for-each (lambda (e) (match e [($ edge/weight u v w) body]))
                 (edges G))]
      [(_ e in G body)
       (for-each (lambda (e) body)
                 (edges G))]))
  
  
  (define (edges G)
    ; return a list of all edges in G
    (let ([all '()])
      (for-each-node v in G
                     (for-each-edge e from v in G 
                                    (set! all (cons e all))))
      all))
  
  ; TODO: OLD VERSION USED THIS FILTER IN THE CASE OF UNORIENTED GRAPHS
  ; [(graph-unoriented? G) (filter (match-lambda [(v . w) (<= v w)])
  ;                               all)])))
  
  (define (in-edges G n)
    ; return list of all in-going edges of n
    ; self-edges are not returned
    (filter (cond
              [(graph-oriented? G)   (match-lambda [($ edge v w) (= w n)])]
              [(graph-unoriented? G) (match-lambda [($ edge v w) (= w n) ;(and (not (= v w)) (or (= v n) (= w n)))
                                                                 ])])
            (edges G)))
  
  (define (in-neighbours G n)
    ; return a list of all nodes which have an edge towards n
    (map (match-lambda [($ edge v w) (if (= v n) w v)])
         (in-edges G n)))
  
  (define (color-graph G type color-node color-edge color-node-and-edge . optionals)
    ; TYPE is either 'depth-first (or 'stack) or 'breadth-first (or 'queue)  
    ; The type determines the discipline used by the non-deterministic set.
    (if (eq? type 'breadth-first) (set! type 'queue))
    (if (eq? type 'depth-first)   (set! type 'stack))
    
    ; We use the following transition system:
    
    ;   i)  X->-O  becomes  X=>=X        calls color-node-and-edge
    ;  ii)  X->-X  becomes  X=>=X        calls color-edge
    ; iii) If there are no pink nodes
    ;        O    becomes    X           calls color-node
    
    ; If the graph isn't oriented then read ->- as ---.
    
    ; Not oriented: A pink node is a red node X with at least one white edge X---? .
    ; Oriented:     A pink node is a red node X with at least one *outgoing* white edge X->-? .
    
    ; Notation:    O   white node
    ;              X   red node
    ;             ---  white edge
    ;             ===  red edge
    
    ; Remark: Each time iii) is used, it means that we have started coloring a new 
    ;         connectedness component.
    
    (let ([PPN (ns:make-set type (graph-no-nodes G))]   ; Potentially Pink Nodes  
          [WN  (ms:set (graph-no-nodes G))])            ; White Nodes
      
      (let ([visit-hook (lambda (n) 'skip)])
        (match optionals
          [()            'skip]
          [(visit)       (set! visit-hook visit)])
        
        (do () [(and (ns:empty? PPN) (ms:empty? WN))  (void)]
          (cond
            ; Transition iii)
            [(ns:empty? PPN)  (let ([v (ms:delete-some! WN)])   
                                (color-node v)
                                (ns:insert! PPN v))]
            [else             (let ([v (ns:delete-some! PPN)])
                                (visit-hook v)
                                (for-each-edge (v w) from v in G
                                               (cond
                                                 ; Transition ii)
                                                 [(ns:member? PPN w)      (color-edge v w)]
                                                 ; Transition i)
                                                 [(ms:member? WN w)       (begin
                                                                            (ms:delete! WN w)
                                                                            (color-node-and-edge v w w)
                                                                            (ns:insert! PPN w))]
                                                 [else 'skip              ; (v,w) is already red
                                                       ]))
                                (if (graph-oriented? G)
                                    (for-each-edge (w v) to v in G
                                                   (cond
                                                     ; Transition ii)
                                                     [(ns:member? PPN w)      (color-edge w v)]
                                                     ; Transition i)
                                                     [(ms:member? WN w)       (begin
                                                                                (ms:delete! WN w)
                                                                                (color-node-and-edge w v w)
                                                                                (ns:insert! PPN w))]
                                                     [else  'skip]))))])))))
  
  
  
  (define (connected-components G)
    (let ([components        (make-vector (graph-no-nodes G) 0)]
          [current -1])
      (color-graph G
                   ; color node
                   (lambda (v) 
                     (set! current (add1 current))
                     (vector-set! components v current))
                   ; color edge
                   (lambda (v w)
                     'skip)
                   ; color edge and node
                   (lambda (v w n)
                     (vector-set! components w current)))
      components))
  
  (define (number-depth-first G)
    (let ([numbers  (make-vector (graph-no-nodes G) 0)]
          [i        -1])
      (color-graph G
                   'stack
                   ; color node
                   (lambda (v) 
                     'skip)
                   ; color edge
                   (lambda (v w)
                     'skip)
                   ; color edge and node
                   (lambda (v w n)
                     'skip)
                   ; hook for delete-some
                   (lambda (v)
                     (set! i (add1 i))
                     (vector-set! numbers v i)))
      numbers))
  
  (define (number-breadth-first G)
    (let ([numbers  (make-vector (graph-no-nodes G) #f)]
          [i        -1])
      (color-graph G
                   'queue
                   ; color node
                   (lambda (v) 
                     (set! i (add1 i))
                     (vector-set! numbers v i))
                   ; color edge
                   (lambda (v w)
                     'skip)
                   ; color edge and node
                   (lambda (v w n)
                     (unless (vector-ref numbers w)
                       (set! i (add1 i))
                       (vector-set! numbers w i))))
      numbers))
  
  (define (spanning-tree G type)
    (let ([tree '()])
      (color-graph G
                   type
                   ; color node
                   (lambda (v) 
                     'skip)
                   ; color edge
                   (lambda (v w)
                     'skip)
                   ; color edge and node
                   (lambda (v w n)
                     (set! tree (cons (list v w) tree))))
      tree))
  
  
  
  ;;; TEST
  
  (define (make-unoriented-graph)
    (make-graph #f))
  
  (define (make-oriented-graph)
    (make-graph #t))
  
  (define (edges->unoriented-graph es)
    (let ([G (make-unoriented-graph)])
      (for-each (match-lambda [(u v)   (add-edge G u v)]
                              [(u v w) (add-edge G u v w)])
                es)
      G))
  
  (define (edges->oriented-graph es)
    (let ([G (make-oriented-graph)])
      (for-each (match-lambda [(u v)   (add-edge G u v)]
                              [(u v w) (add-edge G u v w)])
                es)
      G))
  
  (define G1 
    ; [DAIMI FN 57, p. 14]
    ; Example of breadth first numbering of graph
    (edges->unoriented-graph 
     '((1 2) (1 3) (1 4)
             (2 5) (2 6)
             (3 4)
             (4 7)
             (5 8) (5 9)
             (6 7)
             (7 9)
             (8 9))))
  
  (define G2
    (edges->unoriented-graph
     '((0 1) (0 2)
             (1 2) (1 3) (1 4)
             (2 4)
             (3 4)
             
             (5 6) (5 7)
             (6 7))))
  
  (define G-daimi-page-16
    ; Example of minimum spanning tree of unoriented graph
    ; The weight of a minimum spanning tree is 42
    (edges->unoriented-graph
     '((0 1 8) (0 2 5)
               (1 2 10) (1 3 2) (1 4 18)
               (2 3 3) (2 5 16)
               (3 4 12) (3 5 30) (3 6 14)
               (4 6 4)
               (5 6 26))))
  
  (define G-daimi-page-38
    (edges->oriented-graph
     '((0 2 3) (0 3 4)
               (1 0 8) (1 3 6)
               (2 5 9)
               (3 2 7) (3 4 1) (3 5 5)
               (4 1 6) (4 6 3)
               (5 4 2) (5 6 2))))
  
  (define G3
    ; [Cormen et al., p. 542]
    (edges->oriented-graph
     '((0 1) (0 3)
             (1 4)
             (2 4) (2 5)
             (3 1)
             (4 3))))
  
  (define G4-names #(undershorts pants belt shirt tie jacket socks shoes watch))
  
  (define G4
    ; [Cormen et al., p. 530]
    (edges->oriented-graph
     '((0 1) (0 7)
             (1 2) (1 7)
             (2 5)
             (3 2) (3 4)
             (4 5)
             (6 7)
             (8 8))))
  
  (define (display-G4)
    (let ([name (lambda (n) (vector-ref G4-names n))])
      (map (match-lambda 
             [(from to) (display (name from)) (display " -> ") (display (name to)) (newline)])
           '((0 1) (0 7)
                   (1 2) (1 7)
                   (2 5)
                   (3 2) (3 4)
                   (4 5)
                   (6 7)
                   (8 8)))))
  
  (color-graph G4
               'depth-first
               (lambda (n)     (for-each display (list "            node > " n "\n")))
               (lambda (v w)   (for-each display (list "            edge > " (list v w)   "\n")))
               (lambda (v w n) (for-each display (list "            both > " (list v w) " " n "\n")))
               (lambda (n)     (for-each display (list "visit > " n "\n")))
               )
  
  
  
  (let ([name (lambda (n) (vector-ref G4-names n))])
    (color-graph G4
                 'depth-first
                 (lambda (n)   (for-each display (list "            node > " (name n) "\n")))
                 (lambda (v w) (for-each display (list "            edge > " (list (name v) (name w))   "\n")))
                 (lambda (v w n) (for-each display (list "          both > " (list (name v) (name w)) " " (name n) "\n")))
                 (lambda (n)   (for-each display (list "visit > " (name n) "\n")))
                 ))
  
  (define depth-first-search 
    (case-lambda 
      [(G)             (depth-first-search G (lambda (n) (void)))]
      [(G finish-hook) (let* ([no-nodes      (graph-no-nodes G)]
                              [colors        (make-vector no-nodes 'white)]
                              [predecessor   (make-vector no-nodes #f)]
                              [discovery     (make-vector no-nodes #f)]
                              [finish        (make-vector no-nodes #f)]
                              [time          0]
                              [time++        (thunk (set! time (+ time 1)))])
                         (let-syntax ([color!   (syntax-rules () [(_ v c) (vector-set! colors v c)])]
                                      [color    (syntax-rules () [(_ v)   (vector-ref  colors v)])])
                           (letrec ([visit (lambda (u)
                                             (color! u 'grey)
                                             (time++)
                                             (vector-set! discovery u time)
                                             (for-each-edge (u v) from u in G
                                                            (when (eq? (color v) 'white)
                                                              (vector-set! predecessor v u)
                                                              (visit v)))
                                             (color! u 'black)
                                             (time++)
                                             (vector-set! finish u time)
                                             (finish-hook u))])
                             
                             (for-each-node u in G
                                            (if (eq? (color u) 'white)
                                                (visit u)))
                             
                             (values discovery finish predecessor))))]))
  
  (define (topological-sort G-dag)
    (let* ([sorted       '()]
           [finish-hook  (lambda (n)
                           (set! sorted (cons n sorted)))])
      (depth-first-search G-dag finish-hook)
      sorted))
  
  (define (acyclic? G)
    ; An oriented graph is acyclic, if it doesn't
    ; contain any back edges.
    
    (when (not (graph-oriented? G))
      (error "acyclic?: Expects oriented graph; given " G))
    (let-values ([(d f p) (depth-first-search G)])
      (let ([back-edge? (lambda (u v)
                          (< (vector-ref d v)
                             (vector-ref d u)
                             (vector-ref f u)
                             (vector-ref f v)))])
        (not (ormap (match-lambda 
                      [($ edge u v) (back-edge? u v)])
                    (edges G))))))
  
  
  
  (define (single-source-minimum-path G source)
    ;;;  T[Dijkstra] in O( (V+E)log(V) )
    (let ([remaining       (ps:make-empty (graph-no-nodes G))]  
          [unseen-nodes    (ms:set (graph-no-nodes G))]         
          [distances       (make-vector (graph-no-nodes G) +inf.0)]
          [paths           (make-vector (graph-no-nodes G) #f)])  
      (let-syntax ([distance              (syntax-rules () [(_ n)     (vector-ref  distances n)])]
                   [distance!             (syntax-rules () [(_ n d)   (vector-set! distances n d)])]
                   [path                  (syntax-rules () [(_ n)     (vector-ref  paths n)])]
                   [path!                 (syntax-rules () [(_ n d)   (vector-set! paths n d)])]
                   [insert-remaining!     (syntax-rules () [(_ n e d) (ps:insert!  remaining n e d)])]
                   [delete-min-remaining! (syntax-rules () [(_)       (ps:delete-min! remaining)])])
        
        (ms:delete! unseen-nodes source)
        (distance! source 0)
        (path! source '())
        (for-each-edge e = (source v d) from source in G
                       (begin
                         (ms:delete! unseen-nodes v)
                         (distance! v d)
                         (path! v (list e))
                         (insert-remaining! v e d)))
        (do () [(ps:empty? remaining) 'done]
          (let-values ([(v e d) (ps:find-min remaining)])
            (delete-min-remaining!)
            (distance! v d)
            (path! v (cons e (path (edge-from e))))
            (for-each-edge e = (v w d) from v in G
                           (cond
                             [(ps:member? remaining w)       (if (> (ps:priority remaining w) (+ (distance v) d))
                                                                 (ps:change-priority! remaining w (+ (distance v) d)))]
                             [(ms:member? unseen-nodes w)    (begin
                                                               (ms:delete! unseen-nodes w)
                                                               (insert-remaining! w e (+ (distance v) d)))]))))
        (values distances 
                (do ([i 0 (+ i 1)])
                  [(= i (graph-no-nodes G)) paths]
                  (path! i (if (path i) (reverse (path i)) #f)))))))
  
  (define (minimum-spanning-tree-Prim G)
    ;;; T[Prim] = O( E log(E) )
    (unless (graph-unoriented? G)
      (error "minimum-spanning-treee: Expects unoriented graph; given: " G))
    
    (let ([remaining       (ps:make-empty (graph-no-nodes G))]  
          [unseen-nodes    (ms:set (graph-no-nodes G))]         
          [T               (make-unoriented-graph)])
      (do () [(and (ps:empty? remaining)
                   (ms:empty? unseen-nodes)) T]
        (cond
          [(ps:empty? remaining)   (begin
                                     (let ([u (ms:delete-some! unseen-nodes)])
                                       ; add node v to the tree T
                                       (for-each-edge e = (u v w) from u in G
                                                      (begin
                                                        (ms:delete! unseen-nodes v)
                                                        (ps:insert! remaining v e w)))))]
          [else           (let-values ([(u e w) (ps:delete-min! remaining)])
                            (add-edge T (edge-from e) (edge-to e) (edge/weight-weight e))
                            (for-each-edge e = (u v w) from u in G
                                           (cond
                                             [(ps:member? remaining v)    (if (> (ps:priority remaining v) w)
                                                                              (begin
                                                                                (ps:change-attribute! remaining v e)
                                                                                (ps:change-priority!  remaining v w)))]
                                             [(ms:member? unseen-nodes v) (begin
                                                                            (ms:delete! unseen-nodes v)
                                                                            (ps:insert! remaining v e w))])))]))))
  
  (define (minimum-spanning-tree-Kruskal G)
    (unless (graph-unoriented? G)
      (error "minimum-spanning-treee: Expects unoriented graph; given: " G))
    
    (let ([remaining     (pq:make-empty)]
          [blue          (es:make (graph-no-nodes G))]
          [T             (make-unoriented-graph)])
      (for-each-edge e = (u v w) in G
                     (pq:insert! remaining e w))
      (do () [(pq:empty? remaining) T]
        (let-values ([(e _) (pq:delete-min! remaining)])
          (match e
            [($ edge/weight u v w)  (cond
                                      [(not (es:equivalent? blue u v))  (begin 
                                                                          (add-edge T u v w)
                                                                          (es:join! blue u v))]
                                      [else                            'skip])])))))
  
  )
  
  